# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of kepmodel.
#
# kepmodel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kepmodel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kepmodel.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from . import tools
from .timeseries import SingleTimeseriesModel


class AstroModel(SingleTimeseriesModel):
  r"""
  Astrometric model.

  Parameters
  ----------
  t : (n,) ndarray
    Times of astrometric measurements.
  s : (n,) ndarray
    Astrometric abscissa.
  cth : (n,) ndarray
    Cosine of the scan angle (GAIA convention).
  sth : (n,) ndarray
    Sine of the scan angle (GAIA convention).
  cov : spleaf.cov.Cov (Optional)
    S+LEAF covariance matrix of the astrometric time series.
    If `cov` is not provided, the S+LEAF terms should be provided as `**kwargs`.
  **kwargs :
    Optional arguments (S+LEAF terms) provided to spleaf.cov.Cov to generate
    the S+LEAF covariance matrix of the astrometric time series.
    This is only used if `cov` is not directly provided.
  """

  def __init__(self,
    t,
    s,
    cth,
    sth,
    angular_keplerian_param=True,
    series_index=[],
    **kwargs):
    super().__init__(t, s, series_index, **kwargs)
    self.cth = cth
    self.sth = sth
    self._angular_keplerian_param = angular_keplerian_param

  def _guess_keplerian_harmfit(self, P):
    res = self.residuals()
    u = self.cov.solveL(res) / self.cov.sqD()
    nu = 2 * np.pi / P
    nut_rad = nu * self.t
    Mt = np.concatenate((np.zeros((8, self.full_n)), self.get_fit_lin_M()))
    Mt[0, self.series_index[0]] = self.cth * np.cos(nut_rad)
    Mt[1, self.series_index[0]] = self.sth * np.cos(nut_rad)
    Mt[2, self.series_index[0]] = self.cth * np.sin(nut_rad)
    Mt[3, self.series_index[0]] = self.sth * np.sin(nut_rad)
    Mt[4, self.series_index[0]] = self.cth * np.cos(2 * nut_rad)
    Mt[5, self.series_index[0]] = self.sth * np.cos(2 * nut_rad)
    Mt[6, self.series_index[0]] = self.cth * np.sin(2 * nut_rad)
    Mt[7, self.series_index[0]] = self.sth * np.sin(2 * nut_rad)
    Nt = np.array([self.cov.solveL(Mk) / self.cov.sqD() for Mk in Mt])
    covx = np.linalg.inv(Nt @ Nt.T)
    x = covx @ Nt @ u
    return (x[:8], covx[:8, :8])

  def _guess_keplerian_other(self, P, e, M0_rad):
    res = self.residuals()
    u = self.cov.solveL(res) / self.cov.sqD()
    Mt = np.concatenate((np.zeros((4, self.full_n)), self.get_fit_lin_M()))
    Mt[:4,
      self.series_index[0]] = tools.design_matrix_ABFG(self.t, P, e, M0_rad,
      self.cth, self.sth)
    Nt = np.array([self.cov.solveL(Mk) / self.cov.sqD() for Mk in Mt])
    covx = np.linalg.inv(Nt @ Nt.T)
    x = covx @ Nt @ u
    return (x, covx)

  def guess_keplerian_param(self, P, emax=0.95):
    r"""
    Guess the Keplerian parameters from the period.

    Parameters
    ----------
    P : float
      Period of the Keplerian.
    emax : float
      Maximum allowed eccentricity.

    Returns
    -------
    value : (p,) ndarray
      Values of the Keplerian parameters.
    param : list
      List of the Keplerian parameters names.
    """

    x, covx = self._guess_keplerian_harmfit(P)
    e, M0_rad = tools.solve_eM0(x,
      covx,
      np.array([np.arange(k, 8, 2) for k in range(2)]),
      datatype='astro',
      emax=emax)
    x, _ = self._guess_keplerian_other(P, e, M0_rad)
    return (np.concatenate(([P, M0_rad,
      e], x[:4])), ['P', 'M0', 'e', 'TIA', 'TIB', 'TIF', 'TIG'])

  def keplerian_model(self):
    r"""
    Compute the Keplerian part of the model.

    Returns
    -------
    s : (n,) ndarray
      Astrometric abscissa.
    """

    self._kep_s = np.zeros_like(self.t)
    for kep in self.keplerian.values():
      ddelta, dalpha = kep.astro(self.t)
      self._kep_s += self.cth * ddelta + self.sth * dalpha
    self._kep_coef = 1
    if not self._angular_keplerian_param:
      self._kep_coef = self.get_param('lin.plx')
    return (self._kep_coef * self._kep_s)

  def _keplerian_grad(self, grad_res, grad_dict):
    if not self._angular_keplerian_param:
      if 'lin.plx' in self._default_fit_param:
        grad_plx = -self._kep_s.dot(grad_res)
        grad_dict['lin.plx'] += grad_plx
    grad_res = grad_res * self._kep_coef
    grad_ddelta = -self.cth * grad_res
    grad_dalpha = -self.sth * grad_res
    for name in self.keplerian:
      grad_kep = self.keplerian[name].astro_back(grad_ddelta, grad_dalpha)
      for par, grad_par in zip(self.keplerian[name].get_param(), grad_kep):
        key = f'kep.{name}.{par}'
        if key in self._default_fit_param:
          grad_dict[key] += grad_par

  def _perio_phi(self, cosnut, sinnut):
    return ([
      self.cth * cosnut, self.sth * cosnut, self.cth * sinnut,
      self.sth * sinnut
    ])

  def _fap_d(self):
    return (4)

  def _fap_sqla(self, numax):
    W = self.cov.expandInv()[self.series_index[0], self.series_index[0]]
    nuDt = numax * (self.t[None, :] - self.t[:, None])
    sincnuDt = np.sinc(nuDt / np.pi)
    coscnuDt = -np.sinc(nuDt / (2 * np.pi)) * np.sin(nuDt / 2)
    cosDth = self.cth[None, :] * self.cth[:,
      None] + self.sth[None, :] * self.sth[:, None]
    sinDth = self.cth[None, :] * self.sth[:,
      None] - self.sth[None, :] * self.cth[:, None]
    sincA = sincnuDt * cosDth
    sincB = coscnuDt * sinDth
    sincp = sincA + sincB
    sincm = sincA - sincB
    la = []
    for sinc in [sincp, sincm]:
      Wsinc = W * sinc
      Wsinct = Wsinc @ self.t
      q = np.sum(Wsinc)
      s = np.sum(Wsinct)
      r = self.t @ Wsinct
      la.append(r / q - (s / q)**2)
    sqla = np.sqrt(la)
    return (sqla)



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of kepmodel.
#
# kepmodel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kepmodel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kepmodel.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def _solve_eM0_single(x, indices, datatype, emax):
  r"""
  Solve for e, M0 from the amplitudes of the fundamental and first harmonics
  in a time series (rv or astro).
  """

  xi = x[indices]
  dx = np.zeros((4, x.size))
  for i, j in enumerate(indices):
    dx[i, j] = 1
  x2 = xi * xi

  # Fundamental
  ampf2 = x2[0] + x2[1]
  dampf2 = 2 * (xi[0] * dx[0] + xi[1] * dx[1])

  argf = np.arctan2(-xi[1], xi[0])
  dargf = (xi[1] * dx[0] - xi[0] * dx[1]) / ampf2

  # First harmonics
  amph2 = x2[2] + x2[3]
  damph2 = 2 * (xi[2] * dx[2] + xi[3] * dx[3])

  argh = np.arctan2(-xi[3], xi[2])
  dargh = (xi[3] * dx[2] - xi[2] * dx[3]) / amph2

  # Ratio
  amprho = np.sqrt(amph2 / ampf2)
  if datatype == 'astro':
    amprho *= 2
  damprho = amprho / 2 * (damph2 / amph2 - dampf2 / ampf2)

  argrho = argh - argf
  dargrho = dargh - dargf

  # Phase correction for e^3
  phi = 2 * argf - argh
  dphi = 2 * dargf - dargh

  r = (1 - np.cos(2 * phi) / 6) / 4
  dr = np.sin(2 * phi) / 12 * dphi

  # Solve eccentricity: amprho = e (1 - r e^2)
  if amprho > emax * (1 - r * emax**2):
    e = emax
  else:
    h = np.sqrt(3 * r) / 2
    e = np.cos((np.pi + np.arccos(3 * h * amprho)) / 3) / h
  de = (damprho + e**3 * dr) / (1 - 3 * r * e**2)

  # Solve M0
  a = 1 - r * e**2
  da = -e**2 * dr - 2 * r * e * de
  b = np.sin(2 * phi) / 24 * e**2
  db = (np.cos(2 * phi) * e**2 * dphi + np.sin(2 * phi) * e * de) / 12

  M0 = argrho - np.arctan2(b, a)
  dM0 = dargrho - (a * db - b * da) / (a**2 + b**2)

  return (e, M0, de, dM0)


def _prop_weight_mean(y, dydx, Cx):
  r"""
  Propagate the covariance matrix and compute a weighted mean.
  """

  Cy = dydx @ Cx @ dydx.T
  iCy = np.linalg.inv(Cy)
  u = np.ones(y.size)
  var = 1 / (u.T @ iCy @ u)
  mu = var * (u.T @ (iCy @ y))
  return (mu)


def solve_eM0(x, Cx, indices, datatype, emax):
  r"""
  Solve for the eccentricity and mean anomaly at `t = 0`,
  from the amplitudes of the fundamental and the first harmonics in
  one or more time series (radial velocity and/or astrometry).

  Parameters
  ----------
  x : (4 n,) ndarray
    Coefficients obtained after a linear fit of the time series,
    with the :math:`\cos` and :math:`\sin` of the fundamental
    and first harmonics.
  Cx : (4 n, 4n) ndarray
    Covariance matrix of the parameters `x`.
  indices : (n, 4) ndarray
    Indices corresponding to the :math:`\cos` and :math:`\sin` of the fundamental,
    and :math:`\cos` and :math:`\sin` of the first harmonics for each time series.
  datatype : bool or (n,) ndarray
    Type of each time series ('rv' or 'astro')
  emax : double
    Maximum allowed eccentricity.

  Returns
  -------
  e : double
    Eccentricity estimate.
  M0 : double
    Estimate of the mean anomaly at `t = 0`.
  """

  if len(indices.shape) != 2:
    indices = indices.reshape(1, -1)
  n = indices.shape[0]
  if isinstance(datatype, str):
    datatype = np.full(n, datatype)
  if n == 1:
    e, M0, _, _ = _solve_eM0_single(x, indices[0], datatype[0], emax)
    return (e, M0)
  ec = np.empty(n)
  es = np.empty(n)
  decdx = np.empty((n, x.size))
  desdx = np.empty((n, x.size))
  for k in range(n):
    ek, M0k, dek, dM0k = _solve_eM0_single(x, indices[k], datatype[k], emax)
    ck = np.cos(M0k)
    sk = np.sin(M0k)
    ec[k] = ek * ck
    es[k] = ek * sk
    decdx[k] = ck * dek - es[k] * dM0k
    desdx[k] = sk * dek + ec[k] * dM0k
  mu_ec = _prop_weight_mean(ec, decdx, Cx)
  mu_es = _prop_weight_mean(es, desdx, Cx)
  return (np.sqrt(mu_ec**2 + mu_es**2), np.arctan2(mu_es, mu_ec))


def _M2E(M, e, ftol=5e-16, maxiter=10):
  r"""
  Compute eccentric anomaly from mean anomaly (and eccentricity).
  """

  E = M + 0.85 * np.sign(np.sin(M)) * e
  dE = np.array([1.0 + ftol])
  for _ in range(maxiter):
    d3 = e * np.cos(E)
    d2 = e * np.sin(E)
    d1 = 1 - d3
    diff = M - E + d2
    dE = diff / d1
    dE = diff / (d1 + dE * d2 / 2)
    dE = diff / (d1 + dE * (d2 / 2 + dE * d3 / 6))
    E += dE
    if max(abs(dE)) < ftol:
      break
  return (E)


def _E2v(E, e):
  r"""
  Compute true anomaly from eccentric anomaly (and eccentricity).
  """

  v = 2 * np.arctan(np.sqrt((1 + e) / (1 - e)) * np.tan(E / 2))
  return (v)


def _v2E(v, e):
  r"""
  Compute eccentric anomaly from true anomaly (and eccentricity).
  """

  E = 2 * np.arctan(np.sqrt((1 - e) / (1 + e)) * np.tan(v / 2))
  return (E)


def _E2M(E, e):
  r"""
  Compute mean anomaly from eccentric anomaly (and eccentricity).
  """

  M = E - e * np.sin(E)
  return (M)


def design_matrix_Kom(t, P, e, M0):
  r"""
  Design matrix for the linear fit of
  :math:`K\cos(\omega)` and :math:`K\sin(\omega)`.

  Parameters
  ----------
  t : (n,) ndarray
    Times of the measurements.
  P : double
    Orbital period.
  e : double
    Eccentricity
  M0 : double
    Mean anomaly at `t = 0`.

  Returns
  -------
  phi : (2, n) ndarray
    Design matrix.
  """

  M = M0 + 2 * np.pi * t / P
  E = _M2E(M, e)
  v = _E2v(E, e)
  return (np.array([np.cos(v) + e, -np.sin(v)]))


def design_matrix_ABFG(t, P, e, M0, cth, sth):
  r"""
  Design matrix for the linear fit of Thiele-Innes coefficients
  :math:`A`, :math:`B`, :math:`F`, and :math:`G`.

  Parameters
  ----------
  t : (n,) ndarray
    Times of the measurements.
  P : double
    Orbital period.
  e : double
    Eccentricity
  M0 : double
    Mean anomaly at `t = 0`.
  cth, sth : (n,) ndarrays
    :math:`\cos` and :math:`\sin` of the scan angle.

  Returns
  -------
  phi : (4, n) ndarray
    Design matrix.
  """

  M = M0 + 2 * np.pi * t / P
  E = _M2E(M, e)
  X = np.cos(E) - e
  Y = np.sqrt(1 - e * e) * np.sin(E)
  return (np.array([X * cth, X * sth, Y * cth, Y * sth]))


def smooth_timeseries(t, y, kernel, tau):
  r"""
  Apply the specified kernel to smooth the time series.

  Parameters
  ----------
  t : (n,) ndarray
    Times of the measurements.
  y : (n,) ndarray
    Time series values.
  kernel : function
    Kernel function for the smoothing.
  tau : double
    Smoothing time scale.

  Returns
  -------
  z : (n,) ndarray
    Smoothed time series.
  """

  kdef = ~np.isnan(y)
  x = t / tau
  dx = x[:, None] - x[None, kdef]
  w = kernel(dx)
  w /= np.sum(w, axis=1)[:, None]
  z = w @ y[kdef]
  return (z)


def gaussian_kernel(x):
  """
  Gaussian kernel for the :func:`smooth_timeseries` function.

  Parameters
  ----------
  x : ndarray
    Lag renormalized by time scale.

  Returns
  -------
  w : ndarray
    Corresponding weight.
  """
  return (np.exp(-0.5 * x * x))


def box_kernel(x):
  """
  Box kernel for the :func:`smooth_timeseries` function.

  Parameters
  ----------
  x : ndarray
    Lag renormalized by time scale.

  Returns
  -------
  w : ndarray
    Corresponding weight.
  """
  return (np.abs(x) <= 1.0)


def epanechnikov_kernel(x):
  """
  Epanechnikov kernel for the :func:`smooth_timeseries` function.

  Parameters
  ----------
  x : ndarray
    Lag renormalized by time scale.

  Returns
  -------
  w : ndarray
    Corresponding weight.
  """
  return ((np.abs(x) <= 1.0) * (1.0 - x * x))



import pytest
import numpy as np
from kepmodel.astro import AstroModel
from kepmodel.rv import RvModel
from kepmodel.astrorv import AstroRvModel
from kepderiv import Keplerian
from spleaf import term


def test_astrorvmodel(seed=0):
  np.random.seed(seed)
  n_astro = 251
  n_rv = 351

  th = np.random.uniform(0, 2 * np.pi, n_astro)
  cth = np.cos(th)
  sth = np.sin(th)
  delta = np.random.normal(0, 10)
  alpha = np.random.normal(0, 10)
  mud = np.random.normal(0, 10)
  mua = np.random.normal(0, 10)

  plx = 30.0
  plxfac = np.random.normal(0, 1.0, n_astro)

  offset = np.random.normal(0, 10)

  P = 10**np.random.uniform(2, 3)
  astar = 1e-2
  Marg0 = np.random.uniform(0, 2 * np.pi)
  e = np.random.uniform(0, 0.9)
  w = np.random.uniform(0, 2 * np.pi)
  i = np.arccos(np.random.uniform(0, 1))
  bigw = np.random.uniform(0, 2 * np.pi)
  true_param = np.array(
    [delta, alpha, mud, mua, plx, offset, P, astar, Marg0, e, w, i, bigw])

  kep = Keplerian(true_param[6:], ['P', 'as', 'Marg0', 'e', 'w', 'i', 'bigw'])

  t_astro = np.sort(np.random.uniform(0, 2000, n_astro))
  sig_astro = np.random.uniform(0.1, 0.15, n_astro)
  kepd, kepa = kep.astro(t_astro)
  s = cth * (delta + t_astro * mud + plx * kepd) + sth * (alpha + t_astro * mua
    + plx * kepa) + plx * plxfac + np.random.normal(0, sig_astro)

  t_rv = np.sort(np.random.uniform(0, 2000, n_rv))
  sig_rv = np.random.uniform(3.0, 5.0, n_rv)
  rv = kep.rv(t_rv) + offset + np.random.normal(0, sig_rv)

  Pmin = 1.0
  Pmax = 10000
  nfreq = 10000
  nu0 = 2 * np.pi / Pmax
  dnu = (2 * np.pi / Pmin - nu0) / (nfreq - 1)

  astromodel = AstroModel(t_astro,
    s,
    cth,
    sth,
    err=term.Error(sig_astro),
    angular_keplerian_param=False)
  rvmodel = RvModel(t_rv, rv, err=term.Error(sig_rv))
  fullmodel = AstroRvModel(astromodel, rvmodel)
  astromodel.add_lin(cth, 'delta')
  astromodel.add_lin(sth, 'alpha')
  astromodel.add_lin(cth * t_astro, 'mud')
  astromodel.add_lin(sth * t_astro, 'mua')
  astromodel.add_lin(plxfac, 'plx')
  astromodel.fit()
  astromodel.show_param()

  rvmodel.add_lin(np.ones(n_rv), 'offset')
  rvmodel.fit()
  rvmodel.show_param()

  fullmodel.fit()
  fullmodel.show_param()

  nu, power = fullmodel.periodogram(nu0, dnu, nfreq)

  kmax = np.argmax(power)
  Pmax = 2 * np.pi / nu[kmax]
  fap = fullmodel.fap(power[kmax], nu.max())
  print(P, Pmax, power[kmax], fap)
  assert abs(1 / Pmax - 1 / P) < 1 / (t_rv.max() - t_rv.min())
  assert fap < 1e-3
  fullmodel.add_keplerian_from_period(Pmax)
  fullmodel.set_keplerian_param('0',
    ['P', 'as', 'Marg0', 'e', 'w', 'i', 'bigw'])
  fullmodel.fit()
  fullmodel.show_param()

  param, error = fullmodel.get_param_error()
  param[8:] = (param[8:] - true_param[8:] + np.pi) % (2 *
    np.pi) + true_param[8:] - np.pi
  print(true_param)
  print(param)
  print(error)
  print(((true_param - param) / error)**2)
  chi2r = np.mean(((true_param - param) / error)**2)
  assert chi2r < 2.0



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of kepmodel.
#
# kepmodel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kepmodel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kepmodel.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from . import tools
from .timeseries import SingleTimeseriesModel


class RvModel(SingleTimeseriesModel):
  r"""
  Radial velocity model.

  Parameters
  ----------
  t : (n,) ndarray
    Times of radial velocity measurements.
  rv : (n,) ndarray
    Radial velocity values.
  cov : spleaf.cov.Cov (Optional)
    S+LEAF covariance matrix of the radial velocity time series.
    If `cov` is not provided, the S+LEAF terms should be provided as `**kwargs`.
  **kwargs :
    Optional arguments (S+LEAF terms) provided to spleaf.cov.Cov to generate
    the S+LEAF covariance matrix of the radial velocity time series.
    This is only used if `cov` is not directly provided.
  """

  def __init__(self, t, rv, series_index=[], **kwargs):
    super().__init__(t, rv, series_index, **kwargs)

  def _guess_keplerian_harmfit(self, P):
    res = self.residuals()
    u = self.cov.solveL(res) / self.cov.sqD()
    nu = 2 * np.pi / P
    nut_rad = nu * self.t
    Mt = np.concatenate((np.zeros((4, self.full_n)), self.get_fit_lin_M()))
    Mt[0, self.series_index[0]] = np.cos(nut_rad)
    Mt[1, self.series_index[0]] = np.sin(nut_rad)
    Mt[2, self.series_index[0]] = np.cos(2 * nut_rad)
    Mt[3, self.series_index[0]] = np.sin(2 * nut_rad)
    Nt = np.array([self.cov.solveL(Mk) / self.cov.sqD() for Mk in Mt])
    covx = np.linalg.inv(Nt @ Nt.T)
    x = covx @ Nt @ u
    return (x[:4], covx[:4, :4])

  def _guess_keplerian_other(self, P, e, M0_rad):
    res = self.residuals()
    u = self.cov.solveL(res) / self.cov.sqD()
    Mt = np.concatenate((np.zeros((2, self.full_n)), self.get_fit_lin_M()))
    Mt[:2,
      self.series_index[0]] = tools.design_matrix_Kom(self.t, P, e, M0_rad)
    Nt = np.array([self.cov.solveL(Mk) / self.cov.sqD() for Mk in Mt])
    covx = np.linalg.inv(Nt @ Nt.T)
    x = covx @ Nt @ u
    return (x[:2], covx[:2, :2])

  def guess_keplerian_param(self, P, emax=0.95):
    r"""
    Guess the Keplerian parameters from the period.

    Parameters
    ----------
    P : float
      Period of the Keplerian.
    emax : float
      Maximum allowed eccentricity.

    Returns
    -------
    value : (p,) ndarray
      Values of the Keplerian parameters.
    param : list
      List of the Keplerian parameters names.
    """

    x, covx = self._guess_keplerian_harmfit(P)
    e, M0_rad = tools.solve_eM0(x,
      covx,
      np.arange(4),
      datatype='rv',
      emax=emax)
    x, _ = self._guess_keplerian_other(P, e, M0_rad)
    K = np.sqrt(x[0]**2 + x[1]**2)
    omega_rad = np.arctan2(x[1], x[0])
    return (np.array([P, M0_rad, e, K,
      omega_rad]), ['P', 'M0', 'e', 'K', 'omega'])

  def keplerian_model(self, t=None):
    r"""
    Compute the Keplerian part of the model.

    Parameters
    ----------
    t : (m,) ndarray or None
      Times at which to compute the Keplerian model.
      If None, the measurements times are used.

    Returns
    -------
    rv : (m,) ndarray
      Keplerian model time series.
    """

    if t is None:
      t = self.t
    rv = np.zeros_like(t)
    for kep in self.keplerian.values():
      rv += kep.rv(t)
    return (rv)

  def _keplerian_grad(self, grad_res, grad_dict):
    grad_res = -grad_res
    for name in self.keplerian:
      grad_kep = self.keplerian[name].rv_back(grad_res)
      for par, grad_par in zip(self.keplerian[name].get_param(), grad_kep):
        key = f'kep.{name}.{par}'
        if key in self._default_fit_param:
          grad_dict[key] += grad_par

  def _perio_phi(self, cosnut, sinnut):
    return ([cosnut, sinnut])

  def _fap_d(self):
    return (2)

  def _fap_sqla(self, numax):
    W = self.cov.expandInv()[self.series_index[0], self.series_index[0]]
    nuDt = numax * (self.t[None, :] - self.t[:, None])
    sinc = np.sinc(nuDt / np.pi)
    Wsinc = W * sinc
    Wsinct = Wsinc @ self.t
    q = np.sum(Wsinc)
    s = np.sum(Wsinct)
    r = self.t @ Wsinct
    la = r / q - (s / q)**2
    sqla = np.sqrt(la)
    return (np.array([sqla]))



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of kepmodel.
#
# kepmodel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kepmodel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kepmodel.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from . import tools
from .timeseries import MultiTimeseriesModel
import warnings


class AstroRvModel(MultiTimeseriesModel):
  r"""
  Model combining Astrometry and radial velocities.

  Parameters
  ----------
  astro : AstroModel
    Astrometric model.
  rv : RvModel
    Radial velocity model.
  """

  def __init__(self, astro, rv):
    super().__init__(astro=astro, rv=rv)
    self.astro = astro
    self.rv = rv
    # Enforce use of physical parameters for the astrometry
    if self.astro._angular_keplerian_param:
      warnings.warn(
        'AstroRvModel cannot use angular keplerian parameters,\n'
        'switching AstroModel to use physical keplerian parameters',
        UserWarning)
    self.astro._angular_keplerian_param = False

  def _guess_keplerian_harmfit(self, P):
    x_astro, covx_astro = self.astro._guess_keplerian_harmfit(P)
    x_rv, covx_rv = self.rv._guess_keplerian_harmfit(P)
    x = np.concatenate((x_astro, x_rv))
    covx = np.zeros((x.size, x.size))
    covx[:x_astro.size, :x_astro.size] = covx_astro
    covx[x_astro.size:, x_astro.size:] = covx_rv
    return (x, covx)

  def guess_keplerian_param(self,
    P,
    emax=0.95,
    velocity_coef=1731456.8368055555):
    r"""
    Guess the Keplerian parameters from the period.

    Parameters
    ----------
    P : float
      Period of the Keplerian.
    emax : float
      Maximum allowed eccentricity.
    velocity_coef : double
      Coefficient used for the definition of the velocity unit
      as a function of the distance and time units.
      The default value (1731456.8368055555) corresponds to
      AU for the distance, d for the time, and m/s for the velocity
      (following `IAU - Resolution B3 <http://arxiv.org/abs/1510.07674>`_).

    Returns
    -------
    value : (p,) ndarray
      Values of the Keplerian parameters.
    param : list
      List of the Keplerian parameters names.
    """

    x, covx = self._guess_keplerian_harmfit(P)
    e, M0_rad = tools.solve_eM0(x,
      covx,
      np.array([np.arange(k, 8, 2) for k in range(2)] + [np.arange(8, 12)]),
      datatype=['astro', 'astro', 'rv'],
      emax=emax)
    x_astro, covx_astro = self.astro._guess_keplerian_other(P, e, M0_rad)
    x_rv, covx_rv = self.rv._guess_keplerian_other(P, e, M0_rad)

    # Compute U = (as_AU sini)**2 cos(2 omega)
    #         V = (as_AU sini)**2 sin(2 omega)
    # Astro:
    dx_astro = np.zeros((4, x_astro.size))
    for k in range(4):
      dx_astro[k, k] = 1
    xdx_astro = x_astro[:4, None] * dx_astro
    dplx = np.zeros_like(x_astro)
    plx = self.get_param('astro.lin.plx')
    if 'lin.plx' in self.astro._default_fit_param:
      kplx = 4 + [
        par for par in self.astro._default_fit_param if par.startswith('lin.')
      ].index('lin.plx')
      plx += x_astro[kplx]
      dplx[kplx] = 1

    U_astro = (x_astro[0]**2 + x_astro[1]**2 - x_astro[2]**2 -
      x_astro[3]**2) / plx**2
    dU_astro = 2 * (xdx_astro[0] + xdx_astro[1] - xdx_astro[2] -
      xdx_astro[3]) / plx**2 - 2 * U_astro / plx * dplx
    varU_astro = dU_astro @ covx_astro @ dU_astro
    V_astro = -2 * (x_astro[0] * x_astro[2] + x_astro[1] * x_astro[3]) / plx**2
    dV_astro = -2 * (dx_astro[0] * x_astro[2] + dx_astro[1] * x_astro[3] +
      x_astro[0] * dx_astro[2] +
      x_astro[1] * dx_astro[3]) / plx**2 - 2 * V_astro / plx * dplx
    varV_astro = dV_astro @ covx_astro @ dV_astro

    # RV:
    dx_rv = np.zeros((2, x_rv.size))
    for k in range(2):
      dx_rv[k, k] = 1
    coef = (1 - e**2) * (P / (2 * np.pi * velocity_coef))**2
    U_rv = coef * (x_rv[0]**2 - x_rv[1]**2)
    dU_rv = 2 * coef * (x_rv[0] * dx_rv[0] - x_rv[1] * dx_rv[1])
    varU_rv = dU_rv @ covx_rv @ dU_rv
    V_rv = 2 * coef * x_rv[0] * x_rv[1]
    dV_rv = 2 * coef * (x_rv[1] * dx_rv[0] + x_rv[0] * dx_rv[1])
    varV_rv = dV_rv @ covx_rv @ dV_rv

    # Combine U, V estimates
    U = (U_astro * varU_rv + U_rv * varU_astro) / (varU_rv + varU_astro)
    V = (V_astro * varV_rv + V_rv * varV_astro) / (varV_rv + varV_astro)

    # Deduce asini, omega
    asini = (U**2 + V**2)**(1 / 4)
    om = np.arctan2(V, U) / 2
    om_rv = np.arctan2(x_rv[1], x_rv[0])
    om = (om - om_rv + np.pi / 2) % (np.pi) + om_rv - np.pi / 2

    # Compute i, Omega from astro (Popovic & Pavlovic 1995)
    Om = np.arctan2(x_astro[1] - x_astro[2], x_astro[0] + x_astro[3]) - om
    m = x_astro[0] * x_astro[3] - x_astro[1] * x_astro[2]
    k = np.sum(x_astro[:4]**2) / 2
    j = np.sqrt(k**2 - m**2)
    i = np.arccos(m / (k + j))
    return (np.array([P, asini, M0_rad, e, om, i,
      Om]), ['P', 'assini', 'M0', 'e', 'omega', 'i', 'bigomega'])



import pytest
import numpy as np
from kepmodel.rv import RvModel
from kepderiv import Keplerian
from spleaf import term


def test_rvmodel(seed=0):
  np.random.seed(seed)
  n = 351

  offset = np.random.normal(0, 10)
  P = 10**np.random.uniform(0, 3)
  K = 1.0
  la0 = np.random.uniform(0, 2 * np.pi)
  ecosw = np.random.normal(0, 0.1)
  esinw = np.random.normal(0, 0.1)
  true_param = np.array([offset, P, K, la0, ecosw, esinw])

  kep = Keplerian(true_param[1:], ['P', 'K', 'la0', 'ecosw', 'esinw'])

  t = np.sort(np.random.uniform(0, 2000, n))
  sig = np.random.uniform(0.25, 0.5, n)
  rv = kep.rv(t) + offset + np.random.normal(0, sig)

  Pmin = 1.0
  Pmax = 10000
  nfreq = 10000
  nu0 = 2 * np.pi / Pmax
  dnu = (2 * np.pi / Pmin - nu0) / (nfreq - 1)

  rvmodel = RvModel(t, rv, err=term.Error(sig))
  rvmodel.add_lin(np.ones(n), 'offset')
  rvmodel.fit()
  rvmodel.show_param()
  nu, power = rvmodel.periodogram(nu0, dnu, nfreq)

  kmax = np.argmax(power)
  Pmax = 2 * np.pi / nu[kmax]
  fap = rvmodel.fap(power[kmax], nu.max())
  print(P, Pmax, power[kmax], fap)
  assert abs(1 / Pmax - 1 / P) < 1 / (t.max() - t.min())
  assert fap < 1e-3

  rvmodel.add_keplerian_from_period(Pmax)
  rvmodel.set_keplerian_param('0', ['P', 'K', 'la0', 'ecosw', 'esinw'])
  rvmodel.fit()
  param, error = rvmodel.get_param_error()
  param[3] = (param[3] - true_param[3] + np.pi) % (2 *
    np.pi) + true_param[3] - np.pi
  print(true_param)
  print(param)
  print(error)
  print(((true_param - param) / error)**2)
  chi2r = np.mean(((true_param - param) / error)**2)
  assert chi2r < 2.0



import pytest
import numpy as np
from kepmodel.astro import AstroModel
from kepderiv import Keplerian
from spleaf import term


def test_astromodel(seed=0):
  np.random.seed(seed)
  n = 351

  th = np.random.uniform(0, 2 * np.pi, n)
  cth = np.cos(th)
  sth = np.sin(th)
  delta = np.random.normal(0, 10)
  alpha = np.random.normal(0, 10)
  mud = np.random.normal(0, 10)
  mua = np.random.normal(0, 10)

  plx = 30.0
  plxfac = np.random.normal(0, 1.0, n)

  P = 10**np.random.uniform(0, 3)
  e = np.random.uniform(0, 0.9)
  la0 = np.random.uniform(0, 2 * np.pi)
  A = np.random.normal(0, 1.0)
  B = np.random.normal(0, 1.0)
  F = np.random.normal(0, 1.0)
  G = np.random.normal(0, 1.0)
  true_param = np.array([delta, alpha, mud, mua, plx, P, e, la0, A, B, F, G])

  kep = Keplerian(true_param[5:],
    ['P', 'e', 'la0', 'TIA', 'TIB', 'TIF', 'TIG'])

  t = np.sort(np.random.uniform(0, 2000, n))
  sig = np.random.uniform(0.25, 0.5, n)
  kepd, kepa = kep.astro(t)
  s = cth * (delta + t * mud + kepd) + sth * (alpha + t * mua +
    kepa) + plx * plxfac + np.random.normal(0, sig)

  Pmin = 1.0
  Pmax = 10000
  nfreq = 10000
  nu0 = 2 * np.pi / Pmax
  dnu = (2 * np.pi / Pmin - nu0) / (nfreq - 1)

  astromodel = AstroModel(t, s, cth, sth, err=term.Error(sig))
  astromodel.add_lin(cth, 'delta')
  astromodel.add_lin(sth, 'alpha')
  astromodel.add_lin(cth * t, 'mud')
  astromodel.add_lin(sth * t, 'mua')
  astromodel.add_lin(plxfac, 'plx')
  astromodel.fit()
  astromodel.show_param()
  nu, power = astromodel.periodogram(nu0, dnu, nfreq)

  kmax = np.argmax(power)
  Pmax = 2 * np.pi / nu[kmax]
  fap = astromodel.fap(power[kmax], nu.max())
  print(P, Pmax, power[kmax], fap)
  assert abs(1 / Pmax - 1 / P) < 1 / (t.max() - t.min())
  assert fap < 1e-3

  astromodel.add_keplerian_from_period(Pmax)
  astromodel.set_keplerian_param('0',
    ['P', 'e', 'la0', 'TIA', 'TIB', 'TIF', 'TIG'])
  astromodel.fit()
  param, error = astromodel.get_param_error()
  param[7] = (param[7] - true_param[7] + np.pi) % (2 *
    np.pi) + true_param[7] - np.pi
  print(true_param)
  print(param)
  print(error)
  print(((true_param - param) / error)**2)
  chi2r = np.mean(((true_param - param) / error)**2)
  assert chi2r < 2.0



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of kepmodel.
#
# kepmodel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kepmodel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kepmodel.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup
import os

path = os.path.abspath(os.path.dirname(__file__))

info = {}
with open(os.path.join(path, 'kepmodel', '__info__.py'), 'r') as f:
  exec(f.read(), info)
with open('README.rst', 'r', encoding='utf-8') as readme:
  long_description = readme.read()

setup(name=info['__title__'],
  version=info['__version__'],
  author=info['__author__'],
  author_email=info['__author_email__'],
  license=info['__license__'],
  description=info['__description__'],
  long_description=long_description,
  url=info['__url__'],
  packages=['kepmodel'],
  python_requires='>=3.6',
  install_requires=[
  'numpy>=1.16', 'scipy>=1.2', 'kepderiv>=1.0', 'spleaf>=2.0'
  ])



# -*- coding: utf-8 -*-

# Copyright 2021 Jean-Baptiste Delisle
#
# This file is part of kepmodel.
#
# kepmodel is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kepmodel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kepmodel.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.optimize import minimize
from kepderiv import Keplerian
from spleaf.cov import Cov


class _KeplerianSystem(dict):
  r"""
  Keplerian system.

  This hidden class simply adds the `uid` attribute
  to the dict class in order to generate unique
  names for Keplerians.
  """

  def __init__(self):
    super().__init__()
    self.uid = 0


class ParamTranslator:
  r"""
  Parameter translator.

  This class is a template to define parameter translators
  for :func:`KeplerianModel`.
  One should create a child class and customize its methods,
  to define custom parameters and link them with the default
  :func:`KeplerianModel` parameters.
  Then a member of this child class can be passed
  to :func:`KeplerianModel.set_param_translator`.
  """

  def _link(self, model):
    self.model = model

  def get_default(self, custom_param):
    r"""
    Get the list of default parameters from which
    the provided custom parameters can be computed.

    Parameters
    ----------
    custom_param : list
      Names of the custom parameters.

    Returns
    -------
    default_param : list
      Names of the default parameters.
    """

    return (custom_param)

  def compute_custom(self, custom_param, default_value_dict):
    r"""
    Compute the values of the custom parameters
    from the values of the default ones.

    Parameters
    ----------
    custom_param : list
      Names of the custom parameters.
    default_values_dict : dict
      Values of the default parameters.

    Returns
    -------
    custom_value_dict : dict
      Values of the custom parameters.
    """

    return (default_value_dict)

  def compute_default(self, default_param, custom_value_dict):
    r"""
    Compute the values of the default parameters
    from the values of the custom ones.

    Parameters
    ----------
    default_param : list
      Names of the default parameters.
    custom_values_dict : dict
      Values of the custom parameters.

    Returns
    -------
    default_value_dict : dict
      Values of the default parameters.
    """

    return (custom_value_dict)

  def compute_default_back(self, custom_param, default_grad_dict):
    r"""
    Compute the gradient of a function with respect to the custom parameters
    from its gradient with respect to the default parameters.

    Parameters
    ----------
    custom_param : list
      Names of the custom parameters.
    default_grad_dict : dict
      Gradient with respect to the default parameters.

    Returns
    -------
    custom_grad_dict : dict
      Gradient with respect to the custom parameters.
    """

    return (default_grad_dict)


class KeplerianModel:
  r"""
  Generic class for a Keplerian model.
  """

  def __init__(self):
    self.keplerian = _KeplerianSystem()
    self._default_fit_param = []
    self._custom_fit_param = []
    self._param_translator = ParamTranslator()

  @property
  def nkep(self):
    return (len(self.keplerian))

  def _get_fit_param(self):
    return (self._custom_fit_param.copy())

  def _set_fit_param(self, new_fit_param):
    self._custom_fit_param = new_fit_param
    self._default_fit_param = self._param_translator.get_default(new_fit_param)

  fit_param = property(
    lambda self: self._get_fit_param(),
    lambda self, new_fit_param: self._set_fit_param(new_fit_param))

  def set_param_translator(self, param_translator):
    r"""
    Set the parameter translator to be able to use custom parameters.

    Parameters
    ----------
    param_translator : ParamTranslator
      A :func:`kepmodel.timeseries.ParamTranslator` instance defining the custom parameters.
    """

    param_translator._link(self)
    self._param_translator = param_translator

  def add_keplerian(self, value, param, fit=True, name=None, **kwargs):
    r"""
    Add a Keplerian in the system.

    Parameters
    ----------
    value : (p,) ndarray
      The values of the Keplerian parameters listed in `param`.
    param : list
      List of defined parameters for the Keplerian (see `kepderiv` package).
    fit : bool
      Whether the Keplerian parameters should be added to the fit_param list
      for later fit (see :func:`fit`).
    name : str or None
      Name of the Keplerian (optional).
    **kwargs :
      Additional arguments for Keplerian (see `kepderiv` package).
    """

    if name is None:
      name = f'{self.keplerian.uid}'
      while name in self.keplerian:
        self.keplerian.uid += 1
        name = f'{self.keplerian.uid}'
    self.keplerian[name] = Keplerian(value, param, **kwargs)
    if fit:
      self.fit_param += [f'kep.{name}.{par}' for par in param]
    self.keplerian.uid += 1

  def guess_keplerian_param(self, P, **kwargs):
    raise NotImplementedError

  def add_keplerian_from_period(self,
    P,
    fit=True,
    name=None,
    guess_kwargs={},
    **kwargs):
    r"""
    Add a Keplerian in the system guessing
    all the Keplerian parameters from the period
    (see :func:`guess_keplerian_param`).

    Parameters
    ----------
    value : (p,) ndarray
      The values of the Keplerian parameters listed in `param`.
    P : double
      Orbital period.
    fit : bool
      Whether the Keplerian parameters should be added to the fit_param list
      for later fit (see :func:`fit`).
    name : str or None
      Name of the Keplerian (optional).
    guess_kwargs : dict
      Additional arguments for :func:`guess_keplerian_param`.
    **kwargs :
      Additional arguments for Keplerian (see `kepderiv` package).
    """

    value, param = self.guess_keplerian_param(P, **guess_kwargs)
    self.add_keplerian(value, param, fit, name, **kwargs)

  def set_keplerian_param(self, name, param):
    r"""
    Switch the set of defined parameters for a Keplerian.

    Parameters
    ----------
    name : str
      Name of the Keplerian.
    param : list
      List of new parameters for the Keplerian (see `kepderiv` package).
    """

    fit = False
    new_fit_param = self.fit_param
    for par in self.fit_param:
      if par.startswith(f'kep.{name}.'):
        new_fit_param.remove(par)
        fit = True
    self.keplerian[name].set_param(param)
    if fit:
      new_fit_param += [f'kep.{name}.{par}' for par in param]
    self.fit_param = new_fit_param

  def rm_keplerian(self, name):
    r"""
    Remove a Keplerian.

    Parameters
    ----------
    name : str
      Name of the Keplerian.
    """

    self.keplerian.pop(name)
    new_fit_param = self.fit_param
    for par in self.fit_param:
      if par.startswith(f'kep.{name}.'):
        new_fit_param.remove(par)
    self.fit_param = new_fit_param

  def _chi2(self):
    raise NotImplementedError

  def chi2(self, x=None, param=None, backup=True):
    r"""
    Compute the chi2 of the model.

    Parameters
    ----------
    x : (p,) ndarray or None
      Values of the parameters to set before computing the model.
      If None, keep current values.
    param : list or None
      List of parameter names to set.
      If None, use the fit_param list.
    backup : bool
      Whether to reset the parameters to their current values
      after the call to this method.

    Returns
    -------
    chi2 : float
      The model chi2.
    """

    if x is not None:
      if backup:
        x_old = self.get_param(param)
      self.set_param(x, param)
    chi2 = self._chi2()
    if x is not None and backup:
      self.set_param(x_old, param)
    return (chi2)

  def _loglike(self):
    raise NotImplementedError

  def loglike(self, x=None, param=None, backup=True):
    r"""
    Compute the log-likelihood of the model.

    Parameters
    ----------
    x : (p,) ndarray or None
      Values of the parameters to set before computing the model.
      If None, keep current values.
    param : list or None
      List of parameter names to set.
      If None, use the fit_param list.
    backup : bool
      Whether to reset the parameters to their current values
      after the call to this method.

    Returns
    -------
    loglike : float
      The model log-likelihood.
    """

    if x is not None:
      if backup:
        x_old = self.get_param(param)
      self.set_param(x, param)
    ll = self._loglike()
    if x is not None and backup:
      self.set_param(x_old, param)
    return (ll)

  def _func_default_grad(self, func_name):
    raise NotImplementedError

  def _func_grad(self, func_name, x=None, param=None, backup=True):
    r"""
    Gradient of the method `func_name`.
    """
    if x is not None:
      if backup:
        x_old = self.get_param(param)
      self.set_param(x, param)
    default_grad_dict = self._func_default_grad(func_name)
    grad_dict = self._param_translator.compute_default_back(
      self._custom_fit_param, default_grad_dict)
    grad = np.zeros(len(self._custom_fit_param))
    for k, par in enumerate(self._custom_fit_param):
      grad[k] = grad_dict[par]
    if x is not None and backup:
      self.set_param(x_old, param)
    return (grad)

  def chi2_grad(self, x=None, param=None, backup=True):
    r"""
    Gradient of :func:`chi2` with respect to
    the parameters listed in `fit_param`.

    Parameters
    ----------
    x : (p,) ndarray or None
      Value of the parameters listed in `param`.
      If None, the current value is kept.
    param : list or None
      List of defined parameters.
      If None, the list is assumed to be `fit_param`.
    backup : bool
      whether to reset the parameters to their current values
      after the gradient computation.

    Returns
    -------
    grad : (p,) ndarray
      Gradient.
    """

    return (self._func_grad('chi2', x, param, backup))

  def loglike_grad(self, x=None, param=None, backup=True):
    r"""
    Gradient of :func:`loglike` with respect to
    the parameters listed in `fit_param`.

    Parameters
    ----------
    x : (p,) ndarray or None
      Value of the parameters listed in `param`.
      If None, the current value is kept.
    param : list or None
      List of defined parameters.
      If None, the list is assumed to be `fit_param`.
    backup : bool
      whether to reset the parameters to their current values
      after the gradient computation.

    Returns
    -------
    grad : (p,) ndarray
      Gradient.
    """

    return (self._func_grad('loglike', x, param, backup))

  def loglike_hess(self, x=None, param=None, backup=True, step=1e-6):
    r"""
    Hessian matrix of :func:`loglike` with respect to
    the parameters listed in `fit_param`.

    Parameters
    ----------
    x : (p,) ndarray or None
      Value of the parameters listed in `param`.
      If None, the current value is kept.
    param : list or None
      List of defined parameters.
      If None, the list is assumed to be `fit_param`.
    backup : bool
      Whether to reset the parameters to their current values
      after the gradient computation.
    step : double
      Step size for the second order derivative estimates.

    Returns
    -------
    hess : (p, p) ndarray
      Hessian matrix.
    """

    if x is not None and backup:
      x_old = self.get_param(param)
    llj0 = self.loglike_grad(x, param, False)
    x0 = self.get_param(param)
    xb = list(x0)
    nparam = len(x0)
    hess = np.empty((nparam, nparam))
    for k in range(nparam):
      xb[k] += step
      hess[k] = (self.loglike_grad(xb, param, False) - llj0) / step
      xb[k] = x0[k]
    if x is not None and backup:
      self.set_param(x_old, param)
    else:
      self.set_param(x0, param)
    return ((hess + hess.T) / 2)

  def fit(self, method='L-BFGS-B', step_hess=1e-6, **kwargs):
    r"""
    Adjust the parameters listed in `fit_param` to fit the data.

    This method calls the `scipy.optimize.minimize` function.

    Parameters
    ----------
    method : str
      Minimization method (see `scipy.optimize.minimize`)
    step_hess : double
      Step size for the scale estimate (using :func:`loglike_hess`).
    **kwargs :
      Additional arguments for the `scipy.optimize.minimize` function.
    """

    scale = 1 / np.sqrt(np.abs(np.diag(self.loglike_hess(step=step_hess))))
    bounds = []
    if 'bounds' in kwargs:
      bdict = kwargs.pop('bounds')
    else:
      bdict = {}
    for par, sc in zip(self.fit_param, scale):
      if par in bdict:
        bounds.append((bdict[par][0]/sc, bdict[par][1]/sc))
      # eccentricity:
      elif par.startswith('kep.') and par.endswith('.e'):
        bounds.append((0, 0.95 / sc))
      elif par.startswith('kep.') and ('ecos' in par or 'esin' in par):
        bounds.append((-0.95 / sc, 0.95 / sc))
      # Amplitude
      elif par.startswith('kep.') and (par.endswith('.as')
        or par.endswith('.asini') or par.endswith('.assini')
        or par.endswith('.K')):
        bounds.append((0, None))
      # Period
      elif par.startswith('kep.') and (par.endswith('.P')
        or par.endswith('.n')):
        bounds.append((1e-5 / sc, None))
      # Noise
      elif par.endswith('.sig') or par.endswith('.P0') or par.endswith(
          '.Q') or par.endswith('.rho'):
        bounds.append((1e-5 / sc, None))
      else:
        bounds.append((None, None))
    x_old = self.get_param()
    result = minimize(lambda x: -self.loglike(x * scale, backup=False),
      np.array(x_old) / scale,
      jac=lambda x: -self.loglike_grad(x * scale, backup=False) * scale,
      method=method,
      bounds=bounds,
      **kwargs)
    if result.success:
      self.set_param(result.x * scale)
      self.clean_param()
    else:
      print(result)
      print()
      self.set_param(x_old)
      raise Exception('Fit did not converge.')

  def _chi2ogram(self, nu0, dnu, nfreq):
    raise NotImplementedError

  def periodogram(self, nu0, dnu, nfreq):
    r"""
    Periodogram of the time series.

    Parameters
    ----------
    nu0 : double
      Minimum angular frequency.
    dnu : double
      Step size for the angular frequency.
    nfreq : int
      Number of sampled frequencies.

    Returns
    -------
    nu : (nfreq,) ndarray
      Angular frequencies.
    power : (nfreq,) ndarray
      Periodogram power.
    """

    chi20, chi2 = self._chi2ogram(nu0, dnu, nfreq)
    power = 1.0 - chi2 / chi20
    return (nu0 + np.arange(nfreq) * dnu, power)

  def _fap_Nh(self):
    raise NotImplementedError

  def _fap_d(self):
    raise NotImplementedError

  def _fap_sqla(self, numax):
    raise NotImplementedError

  def Teff(self, numax, d):
    r"""
    Effective time series length for the analytical FAP computation (see :func:`fap`).

    Parameters
    ----------
    numax : double
      Maximum angular frequency sampled in the periodogram.
    d : int
      Number of independant vectors for each frequency.

    Returns
    -------
    Teff : float
      Effective time series length.
    """

    sqla = self._fap_sqla(numax)
    if d == 2:
      return (2 * np.sqrt(np.pi) * sqla[0])
    elif d == 4:
      return (4 / 3 * np.sqrt(np.pi) * (sqla[0] + sqla[1] - sqla[0] * sqla[1] /
        (sqla[0] + sqla[1])))
    elif d == 6:
      sumprod = np.sum([sqla[i] * sqla[j] for i in range(3) for j in range(i)])
      prodsum = np.prod(
        [sqla[i] + sqla[j] for i in range(3) for j in range(i)])
      return (8 / 15 * np.sqrt(np.pi) * (np.sum(sqla) - sumprod**2 / prodsum))
    else:
      raise NotImplementedError

  def fap(self, zmax, numax, Teff=None):
    r"""
    Periodogram False Alarm Probability.

    Parameters
    ----------
    zmax : double
      Power of the highest peak in the periodogram.
    numax : double
      Maximum angular frequency sampled in the periodogram.
    Teff : double or None
      Effective time series length.
      If None, it is automatically computed using :func:`Teff`.

    Returns
    -------
    FAP : double
      Periodogram FAP.
    """

    Nh = self._fap_Nh()
    d = self._fap_d()
    Nk = Nh - d
    fmax = numax / (2.0 * np.pi)
    if Teff is None:
      Teff = self.Teff(numax, d)
    W = fmax * Teff
    chi2ratio = 1.0 - zmax
    FapSingle = chi2ratio**(Nk / 2.0) * (1.0 + Nk / 2.0 * zmax)
    tau = W * np.sqrt(np.pi) * (zmax * Nh /
      (2.0 * np.pi))**((d - 1) / 2) * chi2ratio**(Nk / 2.0 - 1)
    Fap = FapSingle + tau
    if Fap > 1e-5:
      Fap = 1.0 - (1.0 - FapSingle) * np.exp(-tau)
    return (Fap)

  def _get_default_param(self, param):
    raise NotImplementedError

  def get_param(self, param=None):
    r"""
    Get the current values of the model parameters.

    Parameters
    ----------
    param : list or None
      List of parameter names.
      If None, use the fit_param list.

    Returns
    -------
    value : (p,) ndarray
      The values of required parameters.
    """

    singlevar = False
    if param is None:
      param = self.fit_param
    elif isinstance(param, str):
      singlevar = True
      param = [param]
    default_param = self._param_translator.get_default(param)
    default_value_dict = self._get_default_param(default_param)
    value_dict = self._param_translator.compute_custom(param, default_value_dict)
    value = np.empty(len(param))
    for kpar, par in enumerate(param):
      value[kpar] = value_dict[par]
    if singlevar:
      return (value[0])
    else:
      return (value)

  def _set_default_param(self, value_dict):
    raise NotImplementedError

  def set_param(self, value, param=None):
    r"""
    Set the values of the model parameters.

    Parameters
    ----------
    value : (p,) ndarray
      The values of specified parameters.
    param : list or None
      List of parameter names.
      If None, use the fit_param list.
    """

    if param is None:
      param = self.fit_param
    elif isinstance(param, str):
      param = [param]
      value = [value]
    value_dict = dict(zip(param, value))
    default_param = self._param_translator.get_default(param)
    default_value_dict = self._param_translator.compute_default(
      default_param, value_dict)
    self._set_default_param(default_value_dict)

  def get_param_error(self, param=None, step_hess=1e-6):
    r"""
    Get the current values and errorbars of the model parameters.

    Parameters
    ----------
    param : list or None
      List of parameters to show.
      If None, the list is assumed to be `fit_param`.
    step_hess : double
      Step size for the errobars estimates (using :func:`loglike_hess`).
    """

    if param is None:
      param = self.fit_param
    value = self.get_param(param)
    hess = self.loglike_hess(param=param, step=step_hess)
    error = np.sqrt(-np.diag(np.linalg.inv(hess)))
    return (value, error)

  def show_param(self, param=None, step_hess=1e-6, degree=True):
    r"""
    Print a summary of the values and errorbars of the model parameters.

    Parameters
    ----------
    param : list or None
      List of parameters to show.
      If None, the list is assumed to be `fit_param`.
    step_hess : double
      Step size for the errobars estimates (using :func:`loglike_hess`).
    degree : bool
      Whether to print the angles in degrees.
    """

    if param is None:
      param = self.fit_param
    value, error = self.get_param_error(param, step_hess)
    print('{:25s} {:>12s}     {:12s}'.format('Parameter', 'Value', 'Error'))
    for par, val, err in zip(param, value, error):
      parU = par.split('.')[-1].upper().replace('_', '').replace('SQRT',
        'SQ').replace('OMEGA', 'W')
      if degree and parU in [
          'M0', 'MARG0', 'LA0', 'W', 'VARPI', 'VPI', 'BIGW', 'INC', 'I'
      ]:
        par += ' [deg]'
        val *= 180 / np.pi
        err *= 180 / np.pi
      if err == 0 or np.isnan(err):
        dec = 6
      else:
        dec = int(max(0, 3 - np.log10(err)))
      fmt = f'{{:25s}} {{:12.{dec}f}}  ±  {{:<12.{dec}f}}'
      print(fmt.format(par, val, err))
    print()

  def clean_param(self, param=None):
    r"""
    Clean the model parameters to have all angles in :math:`[0, 2\pi]`,
    and the inclination in :math:`[0, \pi]`.

    Parameters
    ----------
    param : list or None
      List of parameters to show.
      If None, the list is assumed to be `fit_param`.
    """

    if param is None:
      param = self.fit_param
    value = self.get_param(param)
    paramU = [
      par.upper().replace('_', '').replace('SQRT', 'SQ').replace('OMEGA',
      'W').replace('INC', 'I') for par in param
    ]
    for kpar, parU in enumerate(paramU):
      parUend = parU.split('.')[-1]
      if parUend in ['M0', 'MARG0', 'LA0', 'W', 'VARPI', 'VPI', 'BIGW', 'I']:
        value[kpar] = value[kpar] % (2 * np.pi)
        if parUend == 'I' and value[kpar] > np.pi:
          try:
            kbigw = paramU.index(parU.replace('.I', '.BIGW'))
            value[kpar] = 2 * np.pi - value[kpar]
            value[kbigw] = (value[kbigw] + np.pi) % (2 * np.pi)
          except:
            pass
    self.set_param(value, param)


class SingleTimeseriesModel(KeplerianModel):
  r"""
  Generic class for a Keplerian model considering a single time series.
  """

  def __init__(self, t, y, series_index=[], **kwargs):
    super().__init__()

    self._parent = None
    self._name = None

    self.full_t = t
    self.full_y = y
    self.full_n = t.size

    self.series_index = series_index + [slice(self.full_n)]
    self.t = t[self.series_index[0]]
    self.y = y[self.series_index[0]]
    self.n = self.t.size

    if 'cov' in kwargs:
      self.cov = kwargs['cov']
    else:
      self.cov = Cov(t, **kwargs)

    self.nlin = 0
    self._lin_uid = 0
    self._lin_name = []
    self._lin_par = np.empty(0)
    self._lin_M = np.empty((0, self.full_n))

  def _set_fit_param(self, new_fit_param):
    if self._parent is None:
      super()._set_fit_param(new_fit_param)
    else:
      self._parent._set_child_fit_param(self._name, new_fit_param)

  def add_lin(self, derivative, name=None, value=0.0, fit=True, series_id=0):
    r"""
    Add a linear predictor to the model.

    The linear predictor is of the form
    :math:`y = \alpha d`
    where `\alpha` is the amplitude (which can adjusted for)
    and `d` is the derivative of the predictor.

    Parameters
    ----------
    derivative : (n,) ndarray
      Time series of the predictor derivative (`d`).
    name : str or None
      Name of the predictor.
    value : float
      Initial value of the amplitude (`\alpha`).
    fit : bool
      Whether the amplitude should be added to the fit_param.
    """

    full_derivative = np.zeros(self.full_n)
    full_derivative[self.series_index[series_id]] = derivative
    self._lin_M = np.vstack((self._lin_M, full_derivative))
    if name is None:
      name = f'{self._lin_uid}'
      while name in self._lin_name:
        self._lin_uid += 1
        name = f'{self._lin_uid}'
    self._lin_name.append(name)
    self._lin_par = np.concatenate((self._lin_par, [value]))
    if fit:
      self.fit_param += [f'lin.{name}']
    self.nlin += 1
    self._lin_uid += 1

  def rm_lin(self, name):
    r"""
    Remove a linear predictor.

    Parameters
    ----------
    name : str
      Name of the predictor.
    """

    klin = self._lin_name.index(name)
    self._lin_M = np.delete(self._lin_M, klin, 0)
    self._lin_name.pop(klin)
    self._lin_par = np.delete(self._lin_par, klin, 0)
    par = f'lin.{name}'
    fit_param = self.fit_param
    if par in fit_param:
      fit_param.remove(par)
      self.fit_param = fit_param
    self.nlin -= 1

  def get_fit_lin_M(self):
    r"""
    Get the matrix of fitted linear predictors derivatives.
    """

    fit_lin_M = np.empty((0, self.full_n))
    for klin in range(self.nlin):
      if f'lin.{self._lin_name[klin]}' in self._default_fit_param:
        fit_lin_M = np.vstack((fit_lin_M, self._lin_M[klin]))
    return (fit_lin_M)

  def _guess_keplerian_harmfit(self, P, **kwargs):
    raise NotImplementedError

  def _guess_keplerian_other(self, P, e, M0_rad, **kwargs):
    raise NotImplementedError

  def keplerian_model(self):
    raise NotImplementedError

  def model(self, x=None, param=None, backup=True, series_id=-1):
    r"""
    Compute the model time series.

    Parameters
    ----------
    x : (p,) ndarray or None
      Values of the parameters to set before computing the model.
      If None, keep current values.
    param : list or None
      List of parameter names to set.
      If None, use the fit_param list.
    backup : bool
      Whether to reset the parameters to their current values
      after the call to this method.

    Returns
    -------
    y : (n,) ndarray
      The model time series.
    """

    if x is not None:
      if backup:
        x_old = self.get_param(param)
      self.set_param(x, param)
    y = self._lin_par.dot(self._lin_M)
    y[self.series_index[0]] += self.keplerian_model()
    if x is not None and backup:
      self.set_param(x_old, param)
    return (y[self.series_index[series_id]])

  def residuals(self, x=None, param=None, backup=True, series_id=-1):
    r"""
    Compute the residuals time series.

    Parameters
    ----------
    x : (p,) ndarray or None
      Values of the parameters to set before computing the model.
      If None, keep current values.
    param : list or None
      List of parameter names to set.
      If None, use the fit_param list.
    backup : bool
      Whether to reset the parameters to their current values
      after the call to this method.

    Returns
    -------
    y : (m,) ndarray
      The residuals time series.
    """
    return (self.full_y[self.series_index[series_id]] -
      self.model(x=x, param=param, backup=backup, series_id=series_id))

  def _chi2(self):
    return (self.cov.chi2(self.residuals()))

  def _loglike(self):
    return (self.cov.loglike(self.residuals()))

  def _keplerian_grad(self, grad_res, grad):
    raise NotImplementedError

  def _func_default_grad(self, func_name):
    getattr(self, func_name)()
    grad_res, grad_noise = getattr(self.cov, func_name + '_grad')()
    grad_dict = {par: 0.0 for par in self._default_fit_param}
    self._keplerian_grad(grad_res[self.series_index[0]], grad_dict)
    for par in self._default_fit_param:
      par_split = par.split('.', 1)
      if par_split[0] == 'lin':
        grad_dict[par] -= self._lin_M[self._lin_name.index(
          par_split[1])].dot(grad_res)
      elif par_split[0] == 'cov':
        grad_dict[par] += grad_noise[self.cov.param.index(par_split[1])]
    return (grad_dict)

  def _perio_phi(self, cosnut, sinnut):
    raise NotImplementedError

  def _chi2ogram(self, nu0, dnu, nfreq):
    res = self.residuals()
    N0t = np.array([
      self.cov.solveL(M0k) / self.cov.sqD() for M0k in self.get_fit_lin_M()
    ]).reshape(-1, res.size)
    u = self.cov.solveL(res) / self.cov.sqD()
    u2 = np.sum(u * u)
    N0tu = N0t @ u
    chi20 = u2 - N0tu.T @ np.linalg.inv(N0t @ N0t.T) @ N0tu

    chi2 = np.empty(nfreq)
    dnut_rad = dnu * self.t
    cosdnut = np.cos(dnut_rad)
    sindnut = np.sin(dnut_rad)
    nu0t_rad = nu0 * self.t
    cosnut = np.cos(nu0t_rad)
    sinnut = np.sin(nu0t_rad)
    full_phik = np.zeros(self.full_n)
    d = self._fap_d()
    Nt = np.empty((d + N0t.shape[0], self.full_n))
    for k, phik in enumerate(self._perio_phi(cosnut, sinnut)):
      full_phik[self.series_index[0]] = phik
      Nt[k] = self.cov.solveL(full_phik) / self.cov.sqD()
    Nt[d:] = N0t
    Ntu = Nt @ u
    chi2[0] = u2 - Ntu.T @ np.linalg.inv(Nt @ Nt.T) @ Ntu
    for kfreq in range(1, nfreq):
      cosnut, sinnut = cosnut * cosdnut - sinnut * sindnut, sinnut * cosdnut + cosnut * sindnut
      for k, phik in enumerate(self._perio_phi(cosnut, sinnut)):
        full_phik[self.series_index[0]] = phik
        Nt[k] = self.cov.solveL(full_phik) / self.cov.sqD()
        Ntu[k] = Nt[k] @ u
      chi2[kfreq] = u2 - Ntu.T @ np.linalg.inv(Nt @ Nt.T) @ Ntu
    return (chi20, chi2)

  def _fap_Nh(self):
    return (self.full_n - np.sum([
      f'lin.{self._lin_name[klin]}' in self._default_fit_param for klin in range(self.nlin)
    ]))

  def _get_default_param(self, param):
    value_dict = {}
    for par in param:
      par_split = par.split('.')
      if par_split[0] == 'kep':
        kpar = self.keplerian[par_split[1]].get_param().index(par_split[2])
        value_dict[par] = self.keplerian[par_split[1]].get_value()[kpar]
      elif par_split[0] == 'lin':
        value_dict[par] = self._lin_par[self._lin_name.index(par[4:])]
      elif par_split[0] == 'cov':
        value_dict[par] = self.cov.get_param([par[4:]])[0]
      else:
        value_dict[par] = np.nan
    return (value_dict)

  def _set_default_param(self, value_dict):
    noise_param = []
    noise_value = []
    kep_value = {
      name: self.keplerian[name].get_value()
      for name in self.keplerian
    }
    kep_change = []
    for par, val in value_dict.items():
      par_split = par.split('.')
      if par_split[0] == 'kep':
        kpar = self.keplerian[par_split[1]].get_param().index(par_split[2])
        kep_value[par_split[1]][kpar] = val
        kep_change.append(par_split[1])
      elif par_split[0] == 'lin':
        self._lin_par[self._lin_name.index(par[4:])] = val
      elif par_split[0] == 'cov':
        noise_param.append(par[4:])
        noise_value.append(val)
    if len(noise_param) > 0:
      self.cov.set_param(noise_value, noise_param)
    for name in kep_change:
      self.keplerian[name].set_value(kep_value[name])


class MultiTimeseriesModel(KeplerianModel):
  r"""
  Generic class for a Keplerian model considering several time series.
  """

  def __init__(self, **kwargs):
    super().__init__()

    self.timeseries = {}
    for key in kwargs:
      if kwargs[key].nkep != 0:
        raise Exception(
          'MultiTimeseriesModel: Cannot merge models containing planets.')
      self.timeseries[key] = kwargs[key]
      self.timeseries[key]._parent = self
      self.timeseries[key]._name = key
      self.timeseries[key].keplerian = self.keplerian
      self._set_child_fit_param(key, self.timeseries[key]._custom_fit_param)

  def _set_fit_param(self, new_fit_param):
    super()._set_fit_param(new_fit_param)
    new_kep_param = []
    for ts in self.timeseries.values():
      ts._custom_fit_param = []
    for par in self._default_fit_param:
      par_split = par.split('.', 1)
      if par_split[0] == 'kep':
        new_kep_param.append(par)
      else:
        self.timeseries[par_split[0]]._custom_fit_param.append(par_split[1])
    for ts in self.timeseries.values():
      ts._custom_fit_param += new_kep_param
      ts._default_fit_param = ts._param_translator.get_default(
        ts._custom_fit_param)

  def _set_child_fit_param(self, key, new_child_fit_param):
    new_fit_param = [
      par for par in self._custom_fit_param
      if not (par.startswith(f'{key}.') or par.startswith('kep.'))
    ]
    for par in new_child_fit_param:
      if par.startswith('kep.'):
        new_fit_param.append(par)
      else:
        new_fit_param.append(f'{key}.{par}')
    self.fit_param = new_fit_param

  def _guess_keplerian_harmfit(self, P, **kwargs):
    xcovx = [
      ts._guess_keplerian_harmfit(P, **kwargs)
      for ts in self.timeseries.values()
    ]
    x = np.array([xck[0] for xck in xcovx])
    covx = np.zeros((x.size, x.size))
    offset = 0
    for xk, covxk in xcovx:
      covx[offset:offset + xk.size, offset:offset + xk.size] = covxk
      offset += xk.size
    return (x, covx)

  def _guess_keplerian_other(self, P, e, M0_rad, **kwargs):
    xcovx = [
      ts._guess_keplerian_other(P, **kwargs)
      for ts in self.timeseries.values()
    ]
    x = np.array([xck[0] for xck in xcovx])
    covx = np.zeros((x.size, x.size))
    offset = 0
    for xk, covxk in xcovx:
      covx[offset:offset + xk.size, offset:offset + xk.size] = covxk
      offset += xk.size
    return (x, covx)

  def _chi2(self):
    return (np.sum([ts.chi2() for ts in self.timeseries.values()]))

  def _loglike(self):
    return (np.sum([ts.loglike() for ts in self.timeseries.values()]))

  def _func_default_grad(self, func_name):
    grad_dict = {par: 0.0 for par in self._default_fit_param}
    for key in self.timeseries:
      default_grad_dict_ts = self.timeseries[key]._func_default_grad(func_name)
      grad_dict_ts = self.timeseries[key]._param_translator.compute_default_back(
        self.timeseries[key]._custom_fit_param, default_grad_dict_ts)
      for par in grad_dict_ts:
        if par.startswith('kep.'):
          grad_dict[par] += grad_dict_ts[par]
        else:
          grad_dict[f'{key}.{par}'] += grad_dict_ts[par]
    return (grad_dict)

  def _chi2ogram(self, nu0, dnu, nfreq):
    chi20 = 0
    chi2 = np.zeros(nfreq)
    for ts in self.timeseries.values():
      chi20k, chi2k = ts._chi2ogram(nu0, dnu, nfreq)
      chi20 += chi20k
      chi2 += chi2k
    return (chi20, chi2)

  def _fap_Nh(self):
    return (np.sum([ts._fap_Nh() for ts in self.timeseries.values()]))

  def _fap_d(self):
    return (np.sum([ts._fap_d() for ts in self.timeseries.values()]))

  def _fap_sqla(self, numax):
    return (np.concatenate(
      [ts._fap_sqla(numax) for ts in self.timeseries.values()]))

  def _get_default_param(self, param):
    value_dict = {}
    ts_param = {key: [] for key in self.timeseries}
    for par in param:
      par_split = par.split('.')
      if par_split[0] == 'kep':
        kpar = self.keplerian[par_split[1]].get_param().index(par_split[2])
        value_dict[par] = self.keplerian[par_split[1]].get_value()[kpar]
      else:
        ts_param[par_split[0]].append('.'.join(par_split[1:]))
    for key in ts_param:
      if len(ts_param[key]) > 0:
        value = self.timeseries[key].get_param(ts_param[key])
        for par, val in zip(ts_param[key], value):
          value_dict[f'{key}.{par}'] = val
    return (value_dict)

  def _set_default_param(self, value_dict):
    ts_param = {key: [] for key in self.timeseries}
    ts_value = {key: [] for key in self.timeseries}
    kep_value = {
      name: self.keplerian[name].get_value()
      for name in self.keplerian
    }
    kep_change = []
    for par, val in value_dict.items():
      par_split = par.split('.')
      if par_split[0] == 'kep':
        kpar = self.keplerian[par_split[1]].get_param().index(par_split[2])
        kep_value[par_split[1]][kpar] = val
        kep_change.append(par_split[1])
      else:
        ts_param[par_split[0]].append('.'.join(par_split[1:]))
        ts_value[par_split[0]].append(val)
    for key in self.timeseries:
      if len(ts_param[key]) > 0:
        self.timeseries[key].set_param(ts_value[key], ts_param[key])
    for name in kep_change:
      self.keplerian[name].set_value(kep_value[name])
